@@ -1,3 +1,42 @@
+"""
+This module provides functionality for interacting with a Dandi Archive server
+via the REST API.  Interaction begins with the creation of a `DandiAPIClient`
+instance, which can be used to retrieve `RemoteDandiset` objects (representing
+Dandisets on the server) and `BaseRemoteAsset` objects (representing assets
+without any data associating them with their Dandisets).  `RemoteDandiset`
+objects can, in turn, be used to retrieve `RemoteAsset` objects (representing
+assets associated with Dandisets).  Aside from `DandiAPIClient`, none of these
+classes should be instantiated directly by the user.
+
+Example code for printing the metadata of all assets with "two-photon" in their
+``metadata.measurementTechnique[].name`` for the latest published version of
+every Dandiset:
+
+>>> import json
+>>> from pathlib import Path
+>>> from dandi.dandiapi import DandiAPIClient
+>>> with DandiAPIClient.for_dandi_instance("dandi") as client:
+...     for dandiset in client.get_dandisets():
+...         if dandiset.most_recent_published_version is None:
+...             continue
+...         latest_dandiset = dandiset.for_version(
+...             dandiset.most_recent_published_version
+...         )
+...         for asset in latest_dandiset.get_assets():
+...             metadata = asset.get_metadata()
+...             if any(
+...                 mtt is not None and "two-photon" in mtt.name
+...                 for mtt in (metadata.measurementTechnique or [])
+...             ):
+...                 print(json.dumps(metadata.json_dict(), indent=4))
+...                 # Uncomment to also download the asset:
+...                 # asset.download(Path(dandiset.identifier, asset.path))
+...
+{
+    ...
+}
+"""
+
 from concurrent.futures import ThreadPoolExecutor, as_completed
 from datetime import datetime
 import json
@@ -36,13 +75,7 @@ from .consts import (
 )
 from .exceptions import NotFoundError, SchemaVersionError
 from .keyring import keyring_lookup
-from .utils import (
-    USER_AGENT,
-    check_dandi_version,
-    ensure_datetime,
-    is_interactive,
-    try_multiple,
-)
+from .utils import USER_AGENT, check_dandi_version, ensure_datetime, is_interactive
 
 lgr = get_logger()
 
@@ -78,7 +111,7 @@ class RESTFullAPIClient:
         json=None,
         headers=None,
         json_resp=True,
-        retry=None,
+        retry_statuses=(),
         **kwargs,
     ):
         """
@@ -117,13 +150,10 @@ class RESTFullAPIClient:
                 print(resp.headers)  # Dict of headers
 
         :type json_resp: bool
-        :param retry: an optional tenacity `retry` argument for retrying the
-            request method
+        :param retry_statuses: a sequence of HTTP response status codes to
+            retry; 503 will be added to this set
         """
 
-        # Look up the HTTP method we need
-        f = getattr(self.session, method.lower())
-
         url = self.get_url(path)
 
         if headers is None:
@@ -133,30 +163,49 @@ class RESTFullAPIClient:
 
         lgr.debug("%s %s", method.upper(), url)
 
-        # urllib3's ConnectionPool isn't thread-safe, so we sometimes hit
-        # ConnectionErrors on the start of an upload.  Retry when this happens.
-        # Cf. <https://github.com/urllib3/urllib3/issues/951>.
-        doretry = tenacity.retry_if_exception_type(
-            requests.ConnectionError
-        ) | tenacity.retry_if_result(lambda r: r.status_code == 503)
-        if retry is not None:
-            doretry |= retry
-
         try:
-            result = try_multiple(12, doretry, 1.25)(
-                f,
-                url,
-                params=params,
-                data=data,
-                files=files,
-                json=json,
-                headers=headers,
-                **kwargs,
-            )
+            for i, attempt in enumerate(
+                tenacity.Retrying(
+                    wait=tenacity.wait_exponential(exp_base=1.25, multiplier=1.25),
+                    # urllib3's ConnectionPool isn't thread-safe, so we
+                    # sometimes hit ConnectionErrors on the start of an upload.
+                    # Retry when this happens.
+                    # Cf. <https://github.com/urllib3/urllib3/issues/951>.
+                    retry=tenacity.retry_if_exception_type(
+                        (requests.ConnectionError, requests.HTTPError)
+                    ),
+                    stop=tenacity.stop_after_attempt(12),
+                    reraise=True,
+                )
+            ):
+                with attempt:
+                    if attempt.retry_state.attempt_number > 1:
+                        lgr.warning("Retrying %s %s", method.upper(), url)
+                    result = self.session.request(
+                        method,
+                        url,
+                        params=params,
+                        data=data,
+                        files=files,
+                        json=json,
+                        headers=headers,
+                        **kwargs,
+                    )
+                    if result.status_code in [503, *retry_statuses]:
+                        result.raise_for_status()
         except Exception:
             lgr.exception("HTTP connection failed")
             raise
 
+        if i > 0:
+            lgr.info(
+                "%s %s succeeded after %d retr%s",
+                method.upper(),
+                url,
+                i,
+                "y" if i == 1 else "ies",
+            )
+
         lgr.debug("Response: %d", result.status_code)
 
         # If success, return the json object. Otherwise throw an exception.
@@ -374,7 +423,13 @@ class APIBase(BaseModel):
 
 
 class Version(APIBase):
-    """The version information for a Dandiset retrieved from the API"""
+    """
+    The version information for a Dandiset retrieved from the API.
+
+    This class should not be instantiated by end-users directly.  Instead,
+    instances should be retrieved from the appropriate attributes & methods of
+    `RemoteDandiset`.
+    """
 
     identifier: str = Field(alias="version")
     name: str
@@ -390,7 +445,11 @@ class Version(APIBase):
 class RemoteDandiset:
     """
     Representation of a Dandiset (as of a certain version) retrieved from the
-    API
+    API.
+
+    This class should not be instantiated by end-users directly.  Instead,
+    instances should be retrieved from the appropriate attributes & methods of
+    `DandiAPIClient` and `RemoteDandiset`.
     """
 
     def __init__(
@@ -503,18 +562,6 @@ class RemoteDandiset:
             **data,
         )
 
-    def _mkasset_from_metadata(self, metadata: Dict[str, Any]) -> "RemoteAsset":
-        return RemoteAsset(
-            client=self.client,
-            dandiset_id=self.identifier,
-            version_id=self.version_id,
-            identifier=metadata["identifier"],
-            path=metadata["path"],
-            size=metadata["contentSize"],
-            modified=metadata["dateModified"],
-            _metadata=metadata,
-        )
-
     def json_dict(self) -> Dict[str, Any]:
         """
         Convert to a JSONable `dict`, omitting the ``client`` attribute and
@@ -627,8 +674,13 @@ class RemoteDandiset:
             if r["status"] == "Valid":
                 return
             sleep(0.5)
+        # TODO: Improve the presentation of the error messages
+        about = {
+            "asset_validation_errors": r.get("asset_validation_errors"),
+            "version_validation_errors": r.get("version_validation_errors"),
+        }
         raise ValueError(
-            f"Dandiset {self.identifier} is {r['status']}: {r['validation_error']}"
+            f"Dandiset {self.identifier} is {r['status']}: {json.dumps(about, indent=4)}"
         )
 
     def publish(self) -> "RemoteDandiset":
@@ -652,9 +704,10 @@ class RemoteDandiset:
         ID.  If the given asset does not exist, a `requests.HTTPError` is
         raised with a 404 status code.
         """
-        return self._mkasset_from_metadata(
-            self.client.get(f"{self.version_api_path}assets/{asset_id}/")
-        )
+        metadata = self.client.get(f"{self.version_api_path}assets/{asset_id}/")
+        asset = self.get_asset_by_path(metadata["path"])
+        asset._metadata = metadata
+        return asset
 
     def get_assets_with_path_prefix(self, path: str) -> Iterator["RemoteAsset"]:
         """
@@ -809,7 +862,7 @@ class RemoteDandiset:
                         lock = Lock()
                         futures = [
                             executor.submit(
-                                upload_part,
+                                _upload_part,
                                 storage_session=storage,
                                 fp=fp,
                                 lock=lock,
@@ -882,13 +935,21 @@ class RemoteDandiset:
 
 
 class BaseRemoteAsset(APIBase):
+    """
+    Representation of an asset retrieved from the API without associated
+    Dandiset information.
+
+    This class should not be instantiated by end-users directly.  Instead,
+    instances should be retrieved from the appropriate attributes & methods of
+    `DandiAPIClient` and `RemoteDandiset`.
+    """
+
     client: "DandiAPIClient"
 
     #: The asset identifier
     identifier: str = Field(alias="asset_id")
     path: str
     size: int
-    modified: datetime
     #: Metadata supplied at initialization; returned when metadata is requested
     #: instead of performing an API call
     _metadata: Optional[Dict[str, Any]] = PrivateAttr(default_factory=None)
@@ -911,7 +972,6 @@ class BaseRemoteAsset(APIBase):
             identifier=metadata["identifier"],
             path=metadata["path"],
             size=metadata["contentSize"],
-            modified=metadata["dateModified"],
             _metadata=metadata,
         )
 
@@ -1039,7 +1099,11 @@ class BaseRemoteAsset(APIBase):
 class RemoteAsset(BaseRemoteAsset):
     """
     Representation of an asset retrieved from the API with associated Dandiset
-    information
+    information.
+
+    This class should not be instantiated by end-users directly.  Instead,
+    instances should be retrieved from the appropriate attributes & methods of
+    `RemoteDandiset`.
     """
 
     JSON_EXCLUDE = frozenset(["client", "dandiset_id", "version_id"])
@@ -1049,6 +1113,10 @@ class RemoteAsset(BaseRemoteAsset):
     #: The identifier for the version of the Dandiset to which the asset
     #: belongs
     version_id: str
+    #: The date at which the asset was created
+    created: datetime
+    #: The date at which the asset was last modified
+    modified: datetime
 
     @property
     def api_path(self) -> str:
@@ -1092,7 +1160,7 @@ class RemoteAsset(BaseRemoteAsset):
         self.client.delete(self.api_path)
 
 
-def upload_part(storage_session, fp, lock, etagger, asset_path, part):
+def _upload_part(storage_session, fp, lock, etagger, asset_path, part):
     etag_part = etagger.get_part(part["part_number"])
     if part["size"] != etag_part.size:
         raise RuntimeError(
@@ -1119,7 +1187,7 @@ def upload_part(storage_session, fp, lock, etagger, asset_path, part):
         part["upload_url"],
         data=chunk,
         json_resp=False,
-        retry=tenacity.retry_if_result(lambda r: r.status_code == 500),
+        retry_statuses=[500],
     )
     server_etag = r.headers["ETag"].strip('"')
     lgr.debug(
